import { Either } from 'fp-ts/lib/Either';
import { PackageJson } from 'type-fest';
import type { ParsedCommandLine } from 'typescript';
import * as PackageManager from '../package-manager';
export declare const DEFAULT_BUILD_FOLDER_PATH_RELATIVE_TO_PROJECT_ROOT = ".nexus/build";
/**
 * The temporary ts build folder used when bundling is enabled
 *
 * Note: It **should not** be nested in a sub-folder. This might "corrupt" the relative paths of the bundle build.
 */
export declare const TMP_TS_BUILD_FOLDER_PATH_RELATIVE_TO_PROJECT_ROOT = ".tmp_build";
/**
 * The part of layout data resulting from the dynamic file/folder inspection.
 */
export declare type ScanResult = {
    app: {
        exists: true;
        path: string;
    } | {
        exists: false;
        path: null;
    };
    project: {
        name: string;
        isAnonymous: boolean;
    };
    sourceRoot: string;
    projectRoot: string;
    nexusModules: string[];
    tsConfig: {
        content: ParsedCommandLine;
        path: string;
    };
    packageManagerType: PackageManager.PackageManager['type'];
    packageJson: null | {
        dir: string;
        path: string;
        content: PackageJson;
    };
};
declare type OutputInfo = {
    startModuleOutPath: string;
    startModuleInPath: string;
    tsOutputDir: string;
    bundleOutputDir: string | null;
    /**
     * The final path to the start module. When bundling disbaled, same as `startModuleOutPath`.
     */
    startModule: string;
    /**
     * The final output dir. If bundler is enabled then this is `bundleOutputDir`.
     * Otherwise it is `tsOutputDir`.
     *
     * When bundle case, this accounts for the bundle environment, which makes it
     * **DIFFERENT** than the source root. For example:
     *
     * ```
     * <out_root>/node_modules/
     * <out_root>/api/app.ts
     * <out_root>/api/index.ts
     * ```
     */
    root: string;
    /**
     * If bundler is enabled then the final output dir where the **source** is
     * located. Otherwise same as `tsOutputDir`.
     *
     * When bundle case, this is different than `root` because it tells you where
     * the source starts, not the build environment.
     *
     * For example, here `source_root` is `<out_root>/api` becuase the user has
     * set their root dir to `api`:
     *
     * ```
     * <out_root>/node_modules/
     * <out_root>/api/app.ts
     * <out_root>/api/index.ts
     * ```
     *
     * But here, `source_root` is `<out_root>` because the user has set their root
     * dir to `.`:
     *
     * ```
     * <out_source_root>/node_modules/
     * <out_source_root>/app.ts
     * <out_source_root>/index.ts
     * ```
     */
    sourceRoot: string;
};
/**
 * The combination of manual datums the user can specify about the layout plus
 * the dynamic scan results.
 */
export declare type Data = ScanResult & {
    build: OutputInfo;
};
/**
 * Layout represents the important edges of the project to support things like
 * scaffolding, build, and dev against the correct paths.
 */
export declare type Layout = Data & {
    /**
     * Property that aliases all the and only the data properties, makes it
     * easy to e.g. serialize just the data.
     */
    data: Data;
    projectRelative(filePath: string): string;
    projectPath(...subPaths: string[]): string;
    /**
     * Like projectPath but treats absolute paths as passthrough.
     */
    projectPathOrAbsolute(...subPaths: string[]): string;
    sourceRelative(filePath: string): string;
    sourcePath(...subPaths: string[]): string;
    update(options: UpdateableLayoutData): void;
    packageManager: PackageManager.PackageManager;
};
interface UpdateableLayoutData {
    nexusModules?: string[];
}
interface Options {
    /**
     * The place to output the build, relative to project root.
     */
    buildOutputDir?: string;
    /**
     * Path to the nexus entrypoint. Can be absolute or relative.
     */
    entrypointPath?: string;
    /**
     * Directory in which the layout should be performed
     */
    cwd?: string;
    /**
     * Whether the build should be outputted as a bundle
     */
    asBundle?: boolean;
}
/**
 * Perform a layout scan and return results with attached helper functions.
 */
export declare function create(options?: Options): Promise<Either<Error, Layout>>;
/**
 * Create a layout based on received data. Useful for taking in serialized scan
 * data from another process that would be wasteful to re-calculate.
 */
export declare function createFromData(layoutData: Data): Layout;
/**
 * Analyze the user's project files/folders for how conventions are being used
 * and where key modules exist.
 */
export declare function scan(opts?: {
    cwd?: string;
    entrypointPath?: string;
}): Promise<ScanResult>;
/**
 * Find the (optional) app module in the user's project.
 */
export declare function findAppModule(opts: {
    projectRoot: string;
}): string | null;
/**
 * Find the project root directory. This can be different than the source root
 * directory. For example the classic project structure where there is a root
 * `src` folder. `src` folder would be considered the "source root".
 *
 * Project root is considered to be the first package.json found from cwd upward
 * to disk root. If not package.json is found then cwd is taken to be the
 * project root.
 *
 * todo update jsdoc or make it true again
 *
 */
export declare function findProjectDir(): string;
/**
 * Detect whether or not CWD is inside a nexus project. nexus project is
 * defined as there being a package.json in or above CWD with nexus as a
 * direct dependency.
 */
export declare function scanProjectType(opts: {
    cwd: string;
}): Promise<{
    type: 'unknown' | 'new';
} | {
    type: 'NEXUS_project' | 'node_project';
    packageJson: {};
    packageJsonLocation: {
        path: string;
        dir: string;
    };
}>;
export declare function saveDataForChildProcess(layout: Layout): {
    NEXUS_LAYOUT: string;
};
/**
 * Load the layout data from a serialized version stored in the environment. If
 * it is not found then a warning will be logged and it will be recalculated.
 * For this reason the function is async however under normal circumstances it
 * should be as-if sync.
 */
export declare function loadDataFromParentProcess(): Promise<Either<Error, Layout>>;
export declare function findNexusModules(tsConfig: ScanResult['tsConfig'], maybeAppModule: string | null): string[];
export {};
//# sourceMappingURL=layout.d.ts.map