import Chalk from 'chalk';
import { stripIndent } from 'common-tags';
import { isLeft, left, right } from 'fp-ts/lib/Either';
import * as FS from 'fs-jetpack';
import * as Path from 'path';
import * as ts from 'ts-morph';
import { findFile, findFileRecurisvelyUpwardSync } from '../../lib/fs';
import { START_MODULE_NAME } from '../../runtime/start/start-module';
import { rootLogger } from '../nexus-logger';
import * as PackageManager from '../package-manager';
import { createContextualError } from '../utils';
import { readOrScaffoldTsconfig } from './tsconfig';
export const DEFAULT_BUILD_FOLDER_PATH_RELATIVE_TO_PROJECT_ROOT = '.nexus/build';
/**
 * The temporary ts build folder used when bundling is enabled
 *
 * Note: It **should not** be nested in a sub-folder. This might "corrupt" the relative paths of the bundle build.
 */
export const TMP_TS_BUILD_FOLDER_PATH_RELATIVE_TO_PROJECT_ROOT = '.tmp_build';
const log = rootLogger.child('layout');
const optionDefaults = {
    buildOutput: DEFAULT_BUILD_FOLDER_PATH_RELATIVE_TO_PROJECT_ROOT,
};
/**
 * Perform a layout scan and return results with attached helper functions.
 */
export async function create(options) {
    var _a;
    const cwd = (_a = options === null || options === void 0 ? void 0 : options.cwd) !== null && _a !== void 0 ? _a : process.cwd();
    const errNormalizedEntrypoint = normalizeEntrypoint(options === null || options === void 0 ? void 0 : options.entrypointPath, cwd);
    if (isLeft(errNormalizedEntrypoint))
        return errNormalizedEntrypoint;
    const normalizedEntrypoint = errNormalizedEntrypoint.right;
    // TODO lodash merge defaults or something
    const scanResult = await scan({ cwd, entrypointPath: normalizedEntrypoint });
    const buildInfo = getBuildInfo(options === null || options === void 0 ? void 0 : options.buildOutputDir, scanResult, options === null || options === void 0 ? void 0 : options.asBundle);
    log.trace('layout build info', { data: buildInfo });
    const layout = createFromData(Object.assign(Object.assign({}, scanResult), { build: buildInfo }));
    /**
     * Save the created layout in the env
     */
    process.env = Object.assign(Object.assign({}, process.env), saveDataForChildProcess(layout));
    return right(layout);
}
/**
 * Create a layout based on received data. Useful for taking in serialized scan
 * data from another process that would be wasteful to re-calculate.
 */
export function createFromData(layoutData) {
    let layout = Object.assign(Object.assign({}, layoutData), { data: layoutData, projectRelative: Path.relative.bind(null, layoutData.projectRoot), sourceRelative: Path.relative.bind(null, layoutData.sourceRoot), sourcePath(...subPaths) {
            return Path.join(layoutData.sourceRoot, ...subPaths);
        },
        projectPath(...subPaths) {
            const joinedPath = Path.join(...subPaths);
            return Path.join(layoutData.projectRoot, joinedPath);
        },
        projectPathOrAbsolute(...subPaths) {
            const joinedPath = Path.join(...subPaths);
            if (Path.isAbsolute(joinedPath))
                return joinedPath;
            return Path.join(layoutData.projectRoot, joinedPath);
        }, packageManager: PackageManager.createPackageManager(layoutData.packageManagerType, {
            projectRoot: layoutData.projectRoot,
        }), update(options) {
            if (options.nexusModules) {
                layout.nexusModules = options.nexusModules;
                layout.data.nexusModules = options.nexusModules;
            }
        } });
    return layout;
}
/**
 * Analyze the user's project files/folders for how conventions are being used
 * and where key modules exist.
 */
export async function scan(opts) {
    var _a, _b;
    log.trace('starting scan');
    const projectRoot = (_a = opts === null || opts === void 0 ? void 0 : opts.cwd) !== null && _a !== void 0 ? _a : process.cwd();
    const packageManagerType = await PackageManager.detectProjectPackageManager({ projectRoot });
    const maybePackageJson = findPackageJson({ projectRoot });
    const maybeAppModule = (_b = opts === null || opts === void 0 ? void 0 : opts.entrypointPath) !== null && _b !== void 0 ? _b : findAppModule({ projectRoot });
    const tsConfig = await readOrScaffoldTsconfig({
        projectRoot,
    });
    const nexusModules = findNexusModules(tsConfig, maybeAppModule);
    const result = {
        app: maybeAppModule === null
            ? { exists: false, path: maybeAppModule }
            : { exists: true, path: maybeAppModule },
        projectRoot,
        sourceRoot: tsConfig.content.options.rootDir,
        nexusModules,
        project: readProjectInfo(opts),
        tsConfig,
        packageManagerType,
        packageJson: maybePackageJson,
    };
    log.trace('completed scan', { result });
    if (result.app.exists === false && result.nexusModules.length === 0) {
        log.error(checks.no_app_or_schema_modules.explanations.problem);
        log.error(checks.no_app_or_schema_modules.explanations.solution);
        process.exit(1);
    }
    return result;
}
// todo allow user to configure these for their project
const CONVENTIONAL_ENTRYPOINT_MODULE_NAME = 'app';
const CONVENTIONAL_ENTRYPOINT_FILE_NAME = `${CONVENTIONAL_ENTRYPOINT_MODULE_NAME}.ts`;
const checks = {
    no_app_or_schema_modules: {
        code: 'no_app_or_schema_modules',
        // prettier-ignore
        explanations: {
            problem: `We could not find any modules that imports 'nexus' or ${CONVENTIONAL_ENTRYPOINT_FILE_NAME} entrypoint`,
            solution: stripIndent `
        Please do one of the following:

          1. Create a file, import { schema } from 'nexus' and write your GraphQL type definitions in it.
          2. Create an ${Chalk.yellow(CONVENTIONAL_ENTRYPOINT_FILE_NAME)} file.
    `,
        }
    },
};
/**
 * Find the (optional) app module in the user's project.
 */
export function findAppModule(opts) {
    log.trace('looking for app module');
    const path = findFile(CONVENTIONAL_ENTRYPOINT_FILE_NAME, opts);
    log.trace('done looking for app module', { path });
    return path;
}
/**
 * Find the project root directory. This can be different than the source root
 * directory. For example the classic project structure where there is a root
 * `src` folder. `src` folder would be considered the "source root".
 *
 * Project root is considered to be the first package.json found from cwd upward
 * to disk root. If not package.json is found then cwd is taken to be the
 * project root.
 *
 * todo update jsdoc or make it true again
 *
 */
export function findProjectDir() {
    return process.cwd();
}
/**
 * Detect whether or not CWD is inside a nexus project. nexus project is
 * defined as there being a package.json in or above CWD with nexus as a
 * direct dependency.
 */
export async function scanProjectType(opts) {
    var _a;
    const packageJsonLocation = findPackageJsonRecursivelyUpward(opts);
    if (packageJsonLocation === null) {
        if (await isEmptyCWD()) {
            return { type: 'new' };
        }
        return { type: 'unknown' };
    }
    const packageJson = FS.read(packageJsonLocation.path, 'json');
    if ((_a = packageJson === null || packageJson === void 0 ? void 0 : packageJson.dependencies) === null || _a === void 0 ? void 0 : _a['nexus']) {
        return {
            type: 'NEXUS_project',
            packageJson: packageJsonLocation,
            packageJsonLocation: packageJsonLocation,
        };
    }
    return {
        type: 'node_project',
        packageJson: packageJsonLocation,
        packageJsonLocation: packageJsonLocation,
    };
}
/**
 * Check if the CWD is empty of any files or folders.
 * TODO we should make nice exceptions for known meaningless files, like .DS_Store
 */
async function isEmptyCWD() {
    const contents = await FS.listAsync();
    return contents === undefined || contents.length === 0;
}
const ENV_VAR_DATA_NAME = 'NEXUS_LAYOUT';
export function saveDataForChildProcess(layout) {
    return {
        [ENV_VAR_DATA_NAME]: JSON.stringify(layout.data),
    };
}
/**
 * Load the layout data from a serialized version stored in the environment. If
 * it is not found then a warning will be logged and it will be recalculated.
 * For this reason the function is async however under normal circumstances it
 * should be as-if sync.
 */
export async function loadDataFromParentProcess() {
    const savedData = process.env[ENV_VAR_DATA_NAME];
    if (!savedData) {
        log.trace('WARNING an attempt to load saved layout data was made but no serialized data was found in the environment. This may represent a bug. Layout is being re-calculated as a fallback solution. This should result in the same layout data (if not, another probably bug, compounding confusion) but at least adds latentency to user experience.');
        return create({}); // todo no build output...
    }
    else {
        // todo guard against corrupted env data
        return right(createFromData(JSON.parse(savedData)));
    }
}
function readProjectInfo(opts) {
    var _a;
    const localFS = FS.cwd((_a = opts === null || opts === void 0 ? void 0 : opts.cwd) !== null && _a !== void 0 ? _a : process.cwd());
    try {
        const packageJson = require(localFS.path('package.json'));
        if (packageJson.name) {
            return {
                name: packageJson.name,
                isAnonymous: false,
            };
        }
    }
    catch (_b) { }
    return {
        name: 'anonymous',
        isAnonymous: true,
    };
}
/**
 * Find the package.json file path. Looks recursively upward to disk root.
 * Starts looking in CWD If no package.json found along search, returns null.
 */
function findPackageJsonRecursivelyUpward(opts) {
    return findFileRecurisvelyUpwardSync('package.json', opts);
}
/**
 *
 */
function findPackageJson(opts) {
    const packageJsonPath = FS.path(opts.projectRoot, 'package.json');
    try {
        const content = FS.read(packageJsonPath, 'json');
        return {
            content,
            path: packageJsonPath,
            dir: Path.dirname(packageJsonPath),
        };
    }
    catch (_a) {
        return null;
    }
}
function normalizeEntrypoint(entrypoint, cwd) {
    if (!entrypoint) {
        return right(undefined);
    }
    const absoluteEntrypoint = entrypoint.startsWith('/') ? entrypoint : Path.join(cwd, entrypoint);
    if (!absoluteEntrypoint.endsWith('.ts')) {
        const error = createContextualError('Entrypoint must be a .ts file', { path: absoluteEntrypoint });
        return left(error);
    }
    if (!FS.exists(absoluteEntrypoint)) {
        const error = createContextualError('Entrypoint does not exist', { path: absoluteEntrypoint });
        return left(error);
    }
    return right(absoluteEntrypoint);
}
/**
 * Get the relative output build path
 * Precedence: User's input > tsconfig.json's outDir > default
 */
function getBuildOutput(buildOutput, scanResult) {
    var _a;
    const output = (_a = buildOutput !== null && buildOutput !== void 0 ? buildOutput : scanResult.tsConfig.content.options.outDir) !== null && _a !== void 0 ? _a : optionDefaults.buildOutput;
    if (Path.isAbsolute(output)) {
        return output;
    }
    return Path.join(scanResult.projectRoot, output);
}
export function findNexusModules(tsConfig, maybeAppModule) {
    try {
        log.trace('finding nexus modules');
        const project = new ts.Project({
            addFilesFromTsConfig: false,
        });
        tsConfig.content.fileNames.forEach((f) => project.addSourceFileAtPath(f));
        const modules = project
            .getSourceFiles()
            .filter((s) => {
            // Do not add app module to nexus modules
            if (s.getFilePath().toString() === maybeAppModule) {
                return false;
            }
            return s.getImportDeclaration('nexus') !== undefined;
        })
            .map((s) => s.getFilePath().toString());
        log.trace('done finding nexus modules', { modules });
        return modules;
    }
    catch (error) {
        log.error('We could not find your nexus modules', { error });
        return [];
    }
}
function getBuildInfo(buildOutput, scanResult, asBundle) {
    const tsOutputDir = getBuildOutput(buildOutput, scanResult);
    const startModuleInPath = Path.join(scanResult.sourceRoot, START_MODULE_NAME + '.ts');
    const startModuleOutPath = Path.join(tsOutputDir, START_MODULE_NAME + '.js');
    if (!asBundle) {
        return {
            tsOutputDir,
            startModuleInPath,
            startModuleOutPath,
            bundleOutputDir: null,
            startModule: startModuleOutPath,
            root: tsOutputDir,
            sourceRoot: tsOutputDir,
        };
    }
    const tsBuildInfo = getBuildInfo(TMP_TS_BUILD_FOLDER_PATH_RELATIVE_TO_PROJECT_ROOT, scanResult, false);
    const relativeRootDir = Path.relative(scanResult.projectRoot, scanResult.tsConfig.content.options.rootDir);
    const sourceRoot = Path.join(tsOutputDir, relativeRootDir);
    return Object.assign(Object.assign({}, tsBuildInfo), { bundleOutputDir: tsOutputDir, root: tsOutputDir, startModule: Path.join(sourceRoot, START_MODULE_NAME + '.js'), sourceRoot });
}
//# sourceMappingURL=layout.js.map